from mastodon_toot_follower import path_util
import json
from mastodon_toot_follower.app import App


class Conversation:
    def __init__(self, app: App, mastodon_instance: str, toot_id: str):
        self.app = app
        self.mastodon = self.app._get_mastodon(mastodon_instance)

        self.toot_id = toot_id
        self.toot = self.mastodon.status(id=self.toot_id)
        self.toot_file = self.app._toot_cache_dir().joinpath(path_util.escape(self.toot['uri']) + '.json')

        self._replies = None
        self._changed_toots = None
        self._saved = False

    def get_conversation_length(self) -> int:
        return len(self.toots())

    def get_changed_toots_and_update(self):
        """
        Compares the old and current versions of this conversation.
        Saves the current version of the conversation afterwards.

        :return: New and changes toots which are part of this conversation.
        """
        if self.is_new():
            self._changed_toots = []
            self.save()
            return self._changed_toots
        else:
            changed_toots = self.get_changed_toots()
            if len(changed_toots) > 0:
                self.save()
            return changed_toots

    def get_changed_toots(self) -> list:
        """
        Compares the old and current versions of this conversation.

        :return: New and changes toots which are part of this conversation.
        """
        if self._changed_toots is None:
            old_toots = self.load()
            old_toots_dict = {old_toot['id']: old_toot['edited_at']
                              for old_toot in old_toots}
            toots = self.toots()

            self._changed_toots = []
            for toot in toots:
                toot_id_not_found = object()
                old_edited_at = old_toots_dict.get(toot['id'], toot_id_not_found)
                if old_edited_at is toot_id_not_found or toot['edited_at'] != old_edited_at:
                    # New reply or has been edited.
                    self._changed_toots.append(toot)

        return self._changed_toots

    def is_new(self):
        """
        :return: True if this is a new conversation of which no old version exists.
        """
        return not self.toot_file.exists()

    def save(self):
        """
        :return: Saves the current version of the conversation. This overwrites the previously saved version.
        """
        if self._saved:
            return
        self._saved = True
        toots = [{'id': entry['id'], 'edited_at': entry['edited_at']}
                 for entry in self.toots()]
        with self.toot_file.open('w') as f:
            f.write(json.dumps(toots))

    def load(self) -> list:
        """
        :return: Loads an old version of this conversation.
        """
        if self._saved:
            raise ValueError('The old version of this conversation has been overwritten.')
        return json.loads(self.toot_file.read_text())

    def toots(self) -> list:
        """
        :return: List of initial toot and it's replies.
        """
        return [self.toot] + self.replies()

    def replies(self) -> list:
        """
        :return: List of replies to initial toot.
        """
        if self._replies is None:
            self._replies = self.mastodon.status_context(id=self.toot_id)['descendants']
        return self._replies
